
import pprint
from optparse import make_option
from django.apps import apps
from django.conf import settings
from djingles import pretenses
from django.core.management import BaseCommand, CommandError


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument("models", nargs='*', help="Name of the model that you wish to truncate")

    def handle(self, **options):
        if not settings.DEBUG:
            raise CommandError("This command cannot be run in production environment")
        models = options["models"]
        if not models:
            for app in apps.get_app_configs():
                self.truncate_app(app)
        else:
            for name in models:
                if "." in name:
                    app_label, model_name = name.split(".", 1)
                    model = apps.get_model(app_label, model_name)
                    self.truncate_model(model)
                else:
                    app = apps.get_app_config(name)
                    self.truncate_app(app)

    def truncate_app(self, app):
        models = app.get_models()
        map(self.truncate_model, models)

    def truncate_model(self, model):
        model.objects.all().delete()




import random
import os
from os import path
from django.utils.lru_cache import lru_cache


__all__ = ['get_random_file', 'get_random_image', 'get_cached_files']


IMAGE_EXTENSIONS = ("jpg", "png", "gif", "jpeg", "bmp", "tiff", "pnga")


@lru_cache(100)
def collect_files(folders, extensions=None):
    _images = []
    for folder in folders:
        for root, dirs, files in os.walk(folder):
            for f in files:
                _, ext = path.splitext(f)
                ext = ext.strip(".").lower()
                if extensions is None or ext in extensions:
                    filename = path.join(root, f)
                    _images.append(filename)
    return _images


def get_cached_files(folders, extensions):
    if not isinstance(folders, (tuple, list)):
        folders = (folders, )
    if extensions is not None:
        extensions = tuple(extensions)
    return collect_files(tuple(folders), extensions)


def get_random_file(folders, extensions=None):
    return random.choice(get_cached_files(folders, extensions))


def get_random_image(folders):
    return get_random_file(folders, IMAGE_EXTENSIONS)




import functools
import inspect
from collections import OrderedDict
import random

from django.core.exceptions import ImproperlyConfigured, FieldDoesNotExist
from django.db.models.fields import CharField, TextField
from django.http.request import QueryDict
from django import forms
from django.utils.encoding import force_text
from djingles.formatters import Formatter
from djingles.forms.fields import SortField
from djingles import utils


__all__ = ['ActionFormMixin', 'ActionModelForm', 'ActionForm', 'FilterForm',
           'FilterModelForm', 'FilterFormMixin', 'action_model_form_factory']


def _find_decorated_methods(form, attribute):
    for name, func in inspect.getmembers(form, inspect.ismethod):
        if name.startswith("_"):
            continue
        if hasattr(func, attribute):
            yield func


class ActionFormMixin(object):

    confirmation_message = None
    failure_message = None
    success_message = None

    def __init__(self, **kwargs):
        self.context = kwargs.pop("context", {})
        self.action = self.context.get("action", "execute")
        self.context["action"] = self.action
        self.action_method = getattr(self, self.action, self.execute)
        super(ActionFormMixin, self).__init__(**kwargs)
        prepare = getattr(self, 'prepare_%s' % self.action, None)
        if prepare is not None:
            prepare()

    def prepare_if_exists(self, func):
        spec = inspect.getfullargspec(func)
        names = [n for n in spec.args if n != "self"]
        try:
            args = [self.fields[f] for f in names]
        except KeyError:
            pass
        else:
            return func(*args)

    def clean_if_exists(self, func):
        spec = inspect.getfullargspec(func)
        names = [n for n in spec.args if n != "self"]
        try:
            args = [self.cleaned_data[f] for f in names]
            if any(a is None for a in args):
                return
        except KeyError:
            pass
        else:
            return func(*args)

    def __bind_fields(self):
        key = "__bound_with_form"
        for name, field in self.fields.items():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    def process_context(self, func):
        context = self.context.copy()
        if hasattr(self, "cleaned_data") and hasattr(self, "save"):
            instance = self.save(commit=False)
            context["instance"] = instance
        context["data"] = self.cleaned_data
        spec = inspect.getfullargspec(func)
        if spec.varargs:
            raise ImproperlyConfigured("Form action cannot have variable arguments")
        if spec.varkw:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}

    def process_result(self, result):
        return result

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        if self.success_message:
            return self.success_message.format(**self.context)
        return self.success_message

    def get_failure_message(self):
        if self.failure_message:
            return self.failure_message.format(**self.context)
        return self.failure_message

    def get_confirmation_message(self):
        if self.confirmation_message:
            return self.confirmation_message.format(**self.context)
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls))

    @classmethod
    def call(cls, *kwargs):
        form_obj = cls(**kwargs)
        if not form_obj.is_valid():
            return False, form_obj.errors
        else:
            return True, form_obj.result

    def full_clean(self):
        result = None
        self.__bind_fields()
        super(ActionFormMixin, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            try:
                if self.is_bound and not self._errors:
                    method = self.action_method
                    context = self.process_context(method)
                    result = method(**context)
                    result = self.process_result(result)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def select_fields(self, exclude=None, include=None, **extras):
        if exclude is not None:
            for f in exclude:
                self.fields.pop(f, None)
        if include is not None:
            include = set(include)
            for f in list(self.fields.keys()):
                if f not in include:
                    self.fields.pop(f)
        if extras:
            for f in extras:
                self.fields[f] = extras[f]

    def execute(self):
        return {}


class ActionForm(ActionFormMixin, forms.Form):
    pass


class ActionModelForm(ActionFormMixin, forms.ModelForm):

    confirmation_message = "Please confirm that you really wish to {action} this object ?"

    def prepare_delete(self):
        self.select_fields(include=())

    def delete(self):
        self.instance.delete()

    def update(self):
        return self.save()

    def create(self):
        return self.save()

    def clone(self):
        self.instance.id = None
        return self.save()


class FilterFormMixin:

    submit_key = None

    def __init__(self, data=None, *args, **kwargs):
        self.context = kwargs.pop("context", {})
        self.boolean_empty_label = kwargs.pop("boolean_empty_label", "Any")
        self.empty_label = kwargs.pop("empty_label", "--------------")
        unbound = False
        if not data or (self.submit_key and self.submit_key not in data):
            data = None
        super(FilterFormMixin, self).__init__(data=data, *args, **kwargs)
        if self.submit_key:
            self.fields[self.submit_key] = forms.IntegerField(
                widget=forms.HiddenInput,
                required=False
            )
        for key, f in self.fields.items():
            if f.disabled:
                continue
            if isinstance(f, forms.BooleanField):
                field = forms.NullBooleanField(required=False,
                                          label=f.label, help_text=f.help_text)
                self.fields[key] = field
                choices=((1, self.boolean_empty_label),(2, "Yes"), (3, "No"))
                field.widget.choices = choices
            if f.required:
                f.required = False
                if hasattr(f, "choices"):
                    choices = list(f.choices)
                    if choices and choices[0][0]:
                        choices.insert(0, ("", self.empty_label))
                    f.choices = tuple(choices)

    def __bind_fields(self):
        key = "__bound_with_form"
        for name, field in self.fields.items():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    @property
    def sort_field(self):
        for name, f in self.fields.items():
            if isinstance(f, SortField):
                return self[name]

    def generic_filter(self, name, queryset, value, data):
        kwargs = {}
        if isinstance(value, (tuple, list)):
            name = '%s__in' % name
        kwargs[name] = value
        return queryset.filter(**kwargs)

    def perform_filter(self, queryset):
        self.__bind_fields()
        if not self.is_bound:
            data = {k: self.get_initial_for_field(f, k) for k, f in self.fields.items()}
        else:
            if not self.is_valid():
                return self.invalid_queryset(queryset)
            data = self.cleaned_data
        allowed = set(self.fields.keys())
        if hasattr(self, "before_filters"):
            queryset = self.before_filters(queryset, data)
        for name, value in data.items():
            if name not in allowed or value in (None, '', [], ()):
                continue
            field = self.fields[name]
            if hasattr(self, "filter_%s" % name):
                result = getattr(self, "filter_%s" % name)(queryset, value, data)
            elif hasattr(field, "filter_queryset"):
                result = field.filter_queryset(queryset, value, self[name])
            else:
                result = self.generic_filter(name, queryset, value, data)
            if result is not None:
                queryset = result
        if hasattr(self, "after_filters"):
            queryset = self.after_filters(queryset, data)
        return queryset

    def invalid_queryset(self, queryset):
        return queryset.none()


class FilterForm(FilterFormMixin, forms.Form):
    pass


class FilterModelForm(FilterFormMixin, forms.ModelForm):

    def generic_filter(self, name, queryset, value, data):
        meta = queryset.model._meta
        try:
            field = meta.get_field(name)
        except FieldDoesNotExist:
            return super(FilterModelForm, self).generic_filter(name, queryset, value, data)
        else:
            op = ""
            if isinstance(value, (list, dict)):
                op = "in"
            elif isinstance(field, (CharField, TextField)):
                op = "icontains"
        query = "%s__%s" % (name, op) if op else name
        kwargs = {query: value}
        return queryset.filter(**kwargs)


def action_model_form_factory(model_class, include=None, exclude=(), **kwargs):
    name = "%sActionForm" % model_class.__name__
    meta = type("Meta", (), {"include": include, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    return type(name, (ActionModelForm, ), kwargs)




from django.urls import reverse
from django.conf import settings
from django.shortcuts import resolve_url
from djingles import utils


__all__ = ["PermissionRequired", "LoginRequired", "PageExpired"]


class CommonHttpError(Exception):

    description = None

    def __init__(self, message=None):
        self.description = message if message is not None else self.description
        super(Exception, self).__init__(self.description)

    def to_json(self):
        messages = getattr(settings, "HTTP_ERROR_MESSAGES",{})
        class_name = self.__class__.__name__
        return {
            'message': messages.get(class_name, self.description),
            'type': class_name
        }


class PageExpired(CommonHttpError):
    status = 410
    description = "Content has moved"


class Redirect(CommonHttpError):
    permanent = False

    description = "Content has moved"

    def __init__(self, to, message=None, **kwargs):
        super(Redirect, self).__init__(message)
        self.to = to
        self.kwargs = kwargs

    def create_url(self, request):
        url = resolve_url(self.to, **self.kwargs)
        return url


class PermissionRequired(Redirect):
    status_code = 401
    description = "Permission Required"

    def __init__(self, url, message=None):
        super(PermissionRequired, self).__init__(url, message=message)

    def create_url(self, request):
        current_url = request.get_full_path()
        url = utils.url_query_update(self.to, {"next": current_url})
        return url


class LoginRequired(PermissionRequired):

    def __init__(self, message=None):
        url = reverse(settings.LOGIN_URL)
        super(LoginRequired, self).__init__(url, message=message)

    def create_url(self, request):
        current_url = request.get_full_path()
        url = utils.url_query_update(self.to, {"next": current_url})
        return url





import pprint
from django.apps import apps
from django.conf import settings
from djingles import pretenses
from django.core.management import BaseCommand, CommandError


class Command(BaseCommand):

    args = "app_name.model_name"
    help = 'Create data for the specified model'

    def add_arguments(self, parser):
        parser.add_argument("model")
        parser.add_argument("-t", "--total", default=20, type=int, help="Total number of instances to be created")
        parser.add_argument("-p", "--pretense", default=None, help="Pretense to be used for content generation")

    def handle(self, **options):
        if not settings.DEBUG:
            raise CommandError("This command cannot be run in production environment")
        name = options["model"]
        if '.' in name:
            app_label, model_name = name.split(".", 1)
            model = apps.get_model(app_label, model_name)
        else:
            try:
                model = next(m for m in apps.get_models() if m.__name__.lower() == name.lower())
            except StopIteration:
                raise ValueError("No model found: %r" % name)
        pretense = options["pretense"]
        verbose = int(options["verbosity"]) > 1
        callback = self.verbose if verbose else None
        pretenses.generate(model, options['total'], name=pretense, callback=callback)

    def verbose(self, data):
        self.stdout.write(pprint.pformat(data, indent=4))
        self.stdout.write("\n")




from django import forms
from djingles.bootstrap4.forms.widgets import SwitchInput


class InlineForm(forms.Form):

    CHOICES = (
        (1, "One"),
        (2, "Two"),
        (3, "Three"),
        (4, "Four"),
        (5, "Five"),
        (6, "Six")
    )

    search = forms.CharField(max_length=100)
    checkbox = forms.BooleanField(widget=forms.CheckboxInput)
    select = forms.TypedChoiceField(widget=forms.Select, choices=CHOICES)


class VerticalForm(InlineForm):

    failure_message = "Failed to complete operation"
    success_message = "Successfully completed the operation"

    CHOICES = (
        (1, "One"),
        (2, "Two"),
        (3, "Three"),
        (4, "Four"),
        (5, "Five"),
        (6, "Six")
    )

    name = forms.CharField(max_length=100)
    avatar = forms.ImageField()
    attachment = forms.FileField()
    # boolean_with_switch = forms.BooleanField(initial=False, help_text="Some useless advice", widget=SwitchInput)
    boolean_with_checkbox = forms.BooleanField(initial=False, help_text="Some useless advice")
    choices_with_radio = forms.TypedChoiceField(coerce=int, choices=CHOICES, widget=forms.RadioSelect)
    choices_with_select = forms.TypedChoiceField(coerce=int, choices=CHOICES, widget=forms.Select)
    multiple_choices_with_select = forms.TypedMultipleChoiceField(coerce=int, choices=CHOICES,
                                                             widget=forms.SelectMultiple)
    multiple_choices_with_checkbox = forms.TypedMultipleChoiceField(coerce=int, choices=CHOICES,
                                                      widget=forms.CheckboxSelectMultiple)
    age = forms.IntegerField()
    textarea = forms.CharField(widget=forms.Textarea, help_text="Text area is a good thing")
    date = forms.DateField()
    time = forms.TimeField()
    datetime = forms.DateTimeField()




from jinja2.nodes import CallBlock
from jinja2.ext import Extension
from djingles import html

class PreExtension(Extension):
    tags = set(['pre'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(['name:endpre'], drop_needle=True)
        return CallBlock(
            self.call_method('_format_content'),
            [],
            [],
            body
        ).set_lineno(lineno)

    def _format_content(self, caller):
        block = caller()
        block = self._format_whitespace(block)
        return block

    def _format_whitespace(self, block):
        blocks = []
        current = []
        empty = 0
        for ln in block.strip().splitlines():
            ln = ln.strip()
            if not ln:
                empty += 1
            else:
                empty = 0
            if not ln and current:
                content = "<p>%s</p>" % ("<br>".join(current))
                current = []
                blocks.append(content)
            elif ln:
                current.append(ln)
            elif empty <= 2:
                blocks.append("<br>")
        if current:
            content = "<p>%s</p>" % ("<br>".join(current))
            blocks.append(content)
        return "".join(blocks)


class TableExtension(Extension):
    tags = set(("table",))

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(['name:endtable'], drop_needle=True)
        return CallBlock(self.call_method('_process_table'),
                               [], [], body).set_lineno(lineno)

    def _process_table(self, caller):
        block = caller()
        delimiter = "||"
        block = self._format_whitespace(block, delimiter)
        return block

    def _format_row(self, row, delimiter):
        columns = row.split(delimiter)
        return html.tr[[html.td(valign='top')[a] for a in columns]]

    def _format_whitespace(self, block, delimiter):
        blocks = []
        current = []
        empty = 0
        for ln in block.strip().splitlines():
            ln = ln.strip()
            if not ln:
                empty += 1
            else:
                empty = 0
            if not ln and current:
                content = self._format_row("<br>".join(current), delimiter)
                current = []
                blocks.append(content)
            elif ln:
                current.append(ln)
            elif empty <= 2:
                blocks.append("<br>")
        if current:
            content = self._format_row("<br>".join(current), delimiter)
            blocks.append(content)
        root = html.table(style="width:100%")[
            html.tbody[
                blocks
            ]
        ]
        return str(root)



from django.core.exceptions import ValidationError
import contextlib
import pytz.exceptions as pyex
import random
import importlib
from django.conf import settings
from django.db import models
from django.apps import apps
from django.db.utils import IntegrityError
from django.forms.models import model_to_dict
from django.utils.text import camel_case_to_spaces
from . import utils, streams


__all__ = ['register', 'generate']


_processors = {}
_callback = None


def register(pretense, *models):
    if isinstance(pretense, type):
        pretense = pretense()
    for m in models:
        _processors[m] = pretense


def generate(model, n=20, name=None, callback=None):
    global _callback
    _callback = callback
    fac = Factory(model)
    return fac.create_all(n)


def on_create(ins):
    if _callback:
        _callback(model_to_dict(ins))


def boolean():
    return random.choice([True, False])


class IncompleteFactoryError(TypeError):
    pass


@contextlib.contextmanager
def impartial(obj):
    if obj.partial:
        raise IncompleteFactoryError
    obj.partial = True
    yield
    obj.partial = False


def get_file(conf_name, formats=None):
    folder = getattr(settings, conf_name)
    return utils.collect_files(folder, formats)


def get_image_file():
    conf_key = "PRETENSE_IMAGE_DIRS"
    return utils.get_random_image(getattr(settings, conf_key))



class Field(object):

    def __init__(self, field, instance, index):
        self.field = field
        self.instance = instance
        self.index = index

    @property
    def null(self):
        return self.field.null

    @property
    def unique(self):
        return self.field.unique

    @property
    def choices(self):
        choices = getattr(self.field, 'choices', None)
        return list(choices) if choices else choices

    @property
    def primary_key(self):
        return self.field.primary_key

    @property
    def model(self):
        return self.field.related_model

    @property
    def choice(self):
        return random.choice(self.choices)[0]

    def has_choices(self):
        return bool(self.choices)

    def __getattr__(self, item):
        return getattr(self.field, item)

    def set(self, value):
        if self.null and random.choice([True, False]):
            value = None
        setattr(self.instance, self.name, value)

    def __getattr__(self, item):
        return getattr(self.field, item)

    def __repr__(self):
        return "Field(%r)" % self.field


class Factory(object):

    __cache = {}
    __modules = None

    highest_id = 0
    total = 0
    name = None

    def __new__(cls, model):
        if model not in cls.__cache:
            obj = super(Factory, cls).__new__(cls)
            cls.__cache[model] = obj
            obj.model = model
            obj.meta = model._meta
            obj.fields = obj.meta.fields
            Factory.load_pretenses()
        else:
            obj = cls.__cache[model]
        return obj

    def __init__(self, model):
        self.update_stats()

    def update_stats(self):
        self.total = self.model.objects.count()
        self.highest_id = self.model.objects.order_by("id").last().id + 1 if self.total else 0

    @classmethod
    def load_pretenses(cls):
        if cls.__modules is None:
            cls.__modules = []
            for app in apps.get_app_configs():
                name = app.module.__name__
                module_name = "%s.pretenses" % name
                try:
                    mod = importlib.import_module(module_name)
                    cls.__modules.append(mod)
                except ImportError as ex:
                    pass

    @property
    def processors(self):
        classes = [self.model] + list(self.model.__mro__)
        result = []
        for k in classes:
            if k not in _processors:
                continue
            proc = _processors[k]
            result.append(proc)
        result.append(DefaultProcessor(self))
        return result

    def post_save(self, ins):
        method_name = "after_%s_%s" % (self.meta.app_label, self.snake_case(self.model))
        for handler in self.processors:
            func = getattr(handler, method_name, None)
            if callable(func):
                func(ins, FactoryProxy)

    def find_method(self, field, default):
        for handler in self.processors:
            stream = getattr(handler, field.name, None)
            if stream:
                if callable(stream):
                    return stream
                func = getattr(stream, 'next', None)
                if callable(func):
                    return func
            for name in self.method_names(field):
                func = getattr(handler, name, None)
                if callable(func):
                    return func
        return default

    @staticmethod
    def snake_case(cls):
        return camel_case_to_spaces(cls.__name__).lower().replace(" ", "_")

    def method_names(self, field):
        template = "process_%s"
        yield template % self.snake_case(field.__class__)
        for cls in field.__class__.__mro__:
            if issubclass(cls, models.Field):
                yield template % self.snake_case(cls)

    def create(self, index, limit, **kwargs):
        ins = self.model()
        for f in self.fields:
            field = Field(f, ins, index)
            field.total = limit
            name = field.name
            if name in kwargs:
                value = kwargs[name]
            else:
                if field.primary_key:
                    continue
                if field.has_choices():
                    value = field.choice
                else:
                    func = self.find_method(f, self.get_field_value)
                    value = func(field)
            if value is not None:
                field.set(value)
        ins.save()
        for f in self.meta.many_to_many:
            field = Field(f, ins, index)
            func = self.find_method(f, self.get_field_value)
            value = func(field)
            if value is not None:
                field.set(value)
        self.total += 1
        self.highest_id = ins.id
        self.post_save(ins)
        on_create(ins)
        return ins

    def get(self):
        if not self.total:
            self.create_all(10)
        return self.model.objects.order_by("?").first()

    def ensure_total(self, total):
        if self.total < total:
            self.create_all(total - self.total)

    def create_all(self, limit, **kwargs):
        i = 0
        errors = 0
        bulk = []
        while i < limit:
            try:
                ins = self.create(self.highest_id+i, limit, **kwargs)
                i += 1
            except (IntegrityError,
                        ValidationError,
                        pyex.AmbiguousTimeError,
                        pyex.NonExistentTimeError,
                        pyex.InvalidTimeError
                    ) as ex:
                errors += 1
                if errors > limit * 5:
                    raise
            else:
                errors = 0
                bulk.append(ins)
        self.update_stats()
        return bulk

    def get_field_value(self, field):
        return None



class FactoryProxy(object):

    def __init__(self, model):
        self.factory = Factory(model)

    def create(self, limit, **kwargs):
        return self.factory.create_all(limit, **kwargs)


class DefaultProcessor(object):

    def __init__(self, factory):
        self.factory = factory

    def process_float_field(self, field):
        return streams.FloatStream().next(field)

    def process_decimal_field(self, field):
        return streams.DecimalStream().next(field)

    def process_null_boolean_field(self, field):
        return boolean()

    def process_boolean_field(self, field):
        return boolean()

    def process_positive_small_integer_field(self, field):
        return streams.IntegerStream(end=2**6).next(field)

    def process_positive_integer_field(self, field):
        return streams.IntegerStream(end=2**24).next(field)

    def process_small_integer_field(self, field):
        return streams.IntegerStream(end=2**6).next(field)

    def process_big_integer_field(self, field):
        return streams.IntegerStream(end=2*24).next(field)

    def process_integer_field(self, field):
        return streams.IntegerStream(end=2**10).next(field)

    def process_slug_field(self, field):
        return streams.SlugStream().next(field)

    def process_char_field(self, field):
        return streams.SentenceStream().next(field)[: field.max_length]

    def process_text_field(self, field):
        return streams.ParagraphStream().next(field)

    def process_foreign_key(self, field):
        if field.null:
            return None
        index = field.index
        factory = Factory(field.model)
        value = factory.get()
        return value

    def process_date_time_field(self, field):
        return streams.DateTimeStream().next(field)

    def process_date_field(self, field):
        return streams.DateStream().next(field)

    def process_time_field(self, field):
        return streams.TimeStream().next(field)

    def process_username(self, field):
        word = streams.FirstNameStream().next(field)
        return "%s%s" % (word, field.index)

    def process_ip_address_field(self, field):
        return streams.IPAddressStream().next(field)

    def process_generic_ip_address_field(self, field):
        return self.process_ip_address_field(field)

    def process_json_field(self, field):
        return {}

    def process_password(self, field):
        word = streams.FirstNameStream().next(field)
        field.instance.set_password(word)

    def process_point_field(self, field):
        return streams.PointStream().next(field)

    def process_image_field(self, field):
        return streams.ImageStream(settings.GINGER_PRETENSE_IMAGE_DIRS).next(field)




import inspect
import functools

from django.core.exceptions import ObjectDoesNotExist
from django.urls import reverse, path
from django.http.response import Http404, HttpResponse, HttpResponseNotAllowed
from django.shortcuts import redirect
from django.urls.exceptions import NoReverseMatch

from djingles.formatters.models import object_formatter_factory, table_formatter_factory
from djingles.forms import action_model_form_factory
from djingles.views.generic import CommonFormView, CommonTemplateView

from djingles.pagination import Paginator


__all__ = [
           'CommonViewSet',
           'CreateViewSetMixin',
           'CommonModelViewSet',
           'UpdateViewSetMixin',
           'ListViewSetMixin',
           'DeleteViewSetMixin',
           'DetailViewSetMixin',
            'CloneViewSetMixin',
            'detail_view',
            'list_view'
           ]


class _SubView(object):
    detail = False
    regex = None
    methods = None

    def __init__(self, name, **kwargs):
        for k,v in kwargs.items():
            setattr(self, k, v)
        self.name = name


def view(**kwargs):
    methods = kwargs.pop("methods", None)

    def wrapper(func):
        subview = _SubView(name=func.__name__, **kwargs)
        wrapped = func
        if methods:
            def closure(self, request, *args, **kw):
                if methods and request.method not in methods:
                    return HttpResponseNotAllowed(request.method)
                response = func(self, request, *args, **kw)
                return response
            wrapped = functools.update_wrapper(closure, func)

        wrapped.__subview__ = subview
        return func

    return wrapper


def list_view(**kwargs):
    kwargs['detail'] = False
    return view(**kwargs)


def detail_view(**kwargs):
    kwargs['detail'] = True
    return view(**kwargs)


def get_child_views(cls):
    for name, func in inspect.getmembers(cls):
        if callable(func) and hasattr(func, '__subview__'):
            yield func.__subview__


class ViewSetMixin:

    action = None
    base_name = None
    url_name = None
    url_regex = None

    def check_object_permissions(self, obj):
        pass

    def check_permissions(self):
        pass

    def get_template_names(self):
        template_name = self.template_name
        if not isinstance(template_name, (list, tuple)):
            template_name = [template_name]
        return [t.format(action=self.action) for t in template_name]

    def reverse(self, action_name, **kwargs):
        url_name = "%s_%s" % (self.base_name, action_name)
        return reverse(url_name, kwargs=kwargs)

    def get(self, request, *args, **kwargs):
        self.check_permissions()
        return getattr(self, self.action)(request)

    def post(self, request, *args, **kwargs):
        self.check_permissions()
        return getattr(self, self.action)(request)

    def process_response(self, request, response):
        if isinstance(response, dict):
            response = self.render_to_response(self.get_context_data(**response))
        return response

    @classmethod
    def get_action_list(cls):
        for name, func in inspect.getmembers(cls):
            if callable(func) and hasattr(func, '__subview__'):
                yield func.__subview__

    @classmethod
    def as_urls(cls, base_name, url_name=None, **kwargs):
        result = []
        # if url_name is None:
        #     url_name = base_name
        for subview in cls.get_action_list():
            action = subview.name
            segments = [url_name] if url_name else []
            if subview.detail:
                segments.append("<int:object_id>")
            if subview.regex is not None:
                regex = subview.regex
            else:
                regex = action
            if regex:
                segments.append(regex)
            segments.append("")
            view_url_name = "%s_%s" % (base_name, action)
            url_regex = "/".join(segments)
            view_class = path(url_regex, cls.as_view(
                base_name=base_name,
                action=subview.name,
                url_name=view_url_name,
                url_regex=url_regex,
                **kwargs
            ), name=view_url_name)
            result.append(view_class)
        return result


class DetailViewSetMixin(object):

    @view(detail=True, regex="")
    def detail(self, request):
        self.object = self.get_object()
        ctx = {
            self.context_object_key: self.object
        }
        object_formatter = self.get_object_formatter()
        if object_formatter:
            ctx[self.context_formatted_object_key] = object_formatter(self.object, variant="detail")
        context = self.get_context_data(**ctx)
        return self.render_to_response(context)


class CreateViewSetMixin(object):

    @view()
    def create(self, request):
        return self.handle_object_form()


class UpdateViewSetMixin(object):

    @view(detail=True)
    def update(self, request):
        self.object = self.get_object()
        return self.handle_object_form()


class DeleteViewSetMixin(object):

    @view(detail=True)
    def delete(self, request):
        self.object = self.get_object()
        return self.handle_object_form()


class CloneViewSetMixin(object):

    @view(detail=True)
    def clone(self, request):
        self.object = self.get_object()
        self.success_url = self.reverse("list")
        return self.handle_object_action()


class ListViewSetMixin(object):

    context_page_key = 'object_list_page'
    context_order_key = 'order'

    @view(regex="")
    def list(self, request):
        object_list_formatter = self.get_object_list_formatter()

        object_list = self.filter_queryset(self.get_queryset(), formatter_class=object_list_formatter)

        filter_form = self.filter_form

        ctx = {}

        object_list = self.paginate_queryset(object_list)

        ctx[self.context_page_key] = object_list

        if object_list_formatter:
            object_list = object_list_formatter(object_list,
                                                sort_key=self.context_order_key,
                                                sort_field=filter_form.fields.get(self.context_order_key) if filter_form else None)

        ctx[self.context_object_list_key] = object_list

        ctx['filter_form'] = self.filter_form

        return self.render_to_response(self.get_context_data(**ctx))


class CommonViewSet(ViewSetMixin, CommonTemplateView):
    pass


class CommonModelViewSet(ViewSetMixin, CommonFormView):

    filter_class = None

    object_formatter = None
    object_list_formatter = None
    action_fields = None

    paginator = Paginator
    url_object_key = 'object_id'
    context_object_key = 'object'
    context_object_list_key = 'object_list'
    context_formatted_object_key = 'formatted_object'
    object_formatter_fields = None
    object_list_formatter_fields = None
    params_page_key = 'page'
    per_page = None

    OK_BACK = 1
    YES_BACK = 2
    CONFIRM_BACK = 3
    SUBMIT_BACK = 4

    def __init__(self, *args, **kwargs):
        super(CommonModelViewSet, self).__init__(*args, **kwargs)
        self.extra_context = {}
        self.form_context = {}
        self.form_initial = {}
        self.form_instance = None
        self.filter_form = None

    def render_to_response(self, ctx, **response_kwargs):
        self.extra_context.update(ctx)
        return super(CommonModelViewSet, self).render_to_response(self.extra_context, **response_kwargs)

    def get_queryset(self):
        return self.queryset

    def filter_queryset(self, queryset, **kwargs):
        self.filter_form = self.get_filter_form(**kwargs)
        if self.filter_form:
            queryset = self.filter_form.perform_filter(queryset)
        return queryset

    def get_filter_kwargs(self):
        return {
            "initial": self.get_filter_initial(),
            "data": self.request.GET,
            "files": None,
            "context": {"request": self.request, "user": self.user}
        }

    def get_object_formatter(self):
        if self.object_formatter is None:
            model_class = self.get_queryset().model
            self.object_formatter = object_formatter_factory(
                model_class,
                fields=self.object_formatter_fields
            )
        return self.object_formatter

    def get_object_list_formatter(self):
        if self.object_list_formatter is None:
            model_class = self.get_queryset().model
            self.object_list_formatter = table_formatter_factory(
                model_class,
                fields=self.object_list_formatter_fields,
                get_cell_url=lambda me, cell: cell.source.get_absolute_url() if
                        hasattr(cell.source, 'get_absolute_url') else self.reverse("detail", object_id=cell.source.id)
            )
        return self.object_list_formatter

    def get_form_class(self, form_key=None):
        form_class = super(CommonModelViewSet, self).get_form_class(form_key)
        if form_class is None:
            model_class = self.get_queryset().model
            form_class = action_model_form_factory(model_class, include=self.action_fields)
        return form_class

    def get_filter_initial(self):
        return None

    def get_filter_form(self, **extra):
        filter_class = self.get_filter_class()
        if filter_class:
            kwargs = self.get_filter_kwargs()
            kwargs["context"].update(extra)
            return filter_class(**kwargs)

    def paginate_queryset(self, queryset, per_page=None, params_page_key=None):
        if per_page is None:
            per_page = self.per_page
        if params_page_key is None:
            params_page_key = self.params_page_key
        if not per_page:
            return queryset
        return self.paginator(
            queryset,
            per_page=per_page,
            parameter_name=params_page_key,
            allow_empty=False
        ).page(self.request)

    def get_object(self):
        queryset = self.get_queryset()
        try:
            obj = queryset.get(pk=self.kwargs[self.url_object_key])
            self.check_object_permissions(obj)
            return obj
        except queryset.model.DoesNotExist:
            raise Http404

    def get_form_action(self):
        return self.action

    def get_form_context(self, form_key):
        ctx = super(CommonModelViewSet, self).get_form_context(form_key)
        ctx['request'] = self.request
        ctx['action'] = self.get_form_action()
        ctx.update(self.form_context)
        return ctx

    def get_form_initial(self, form_key):
        initial = super(CommonModelViewSet, self).get_form_initial(form_key) or {}
        initial.update(self.form_initial)
        return initial

    def get_form_instance(self, form_key):
        form_instance = self.form_instance
        return form_instance if form_instance is not None else getattr(self, self.context_object_key, None)

    def get_filter_class(self):
        return self.filter_class

    def render_context(self, context):
        return self.render_to_response(self.get_context_data(**context))

    def handle_object_action(self, **kwargs):
        if not hasattr(self, 'object'):
            self.object = self.get_object()
        return self.handle_object_form(**kwargs)

    def handle_object_form(self, **kwargs):
        request = self.request
        method = request.method
        self.extra_context.update(kwargs)
        if method == 'GET':
            form = self.get_form(None, data=None, files=None)
            return self.render_form(form)
        else:
            try:
                if self.success_url is None:
                    if self.action in {'delete', "create"}:
                        self.success_url = self.reverse('list')
                    else:
                        self.success_url = self.reverse('detail', object_id=self.object.id)
            except NoReverseMatch:
                pass
            return self.process_submit(None, data=request.POST, files=request.FILES)

    def handle_queryset_action(self):
        request = self.request
        method = request.method
        if method == 'GET':
            return HttpResponse("Bad Request", status=400)
        else:
            object_list = self.get_queryset()
            form = self.filter_class(queryset=object_list, action=self.action, data=request.GET)
            if form.is_valid():
                return redirect(self.get_success_url(form))
            else:
                return redirect(self.get_previous_url())






from jinja2 import Markup
from django import forms
from djingles import html
from djingles.forms.widgets import AbstractThemedWidget, ChoiceWidgetMixin


class BootstrapWidget(AbstractThemedWidget):

    error_class = "field-errors"
    label_class = "field-label"
    field_class = "form-group"
    help_class = "help-block"
    failure_class = "has-error"
    success_class = "has-success"
    widget_class = ""
    input_class = "form-control"

    def context_for_field(self, field):
        return {
            "field": field,
            "widget_tag": self.render_widget(field),
            "label": field.label,
            "id_for_label": field.id_for_label,
            "label_tag": self.render_label(field),
            "help": field.help_text,
            "help_tag": self.render_help(field),
            "errors": field.errors,
            "error_tag": self.render_errors(field)
        }

    def render_label(self, field):
        if not field.label:
            return ""
        return html.label(class_=self.label_class, for_=field.id_for_label)[
            field.label
        ]

    def render_help(self, field):
        help_text = field.help_text
        if not help_text:
            return ""
        return html.div(class_=self.help_class)[
            field.help_text
            ]

    def render_errors(self, field):
        errors = field.errors
        return html.ul(class_=self.error_class)[
            [html.li[err] for err in errors]
        ]

    def get_form_property(self, field, property_name, default=None):
        form = field.form
        prop = getattr(form, property_name, None)
        if prop is not None:
            if callable(prop):
                return prop(field.name)
            else:
                return prop.get(field.name)

    def wrap_widget(self, field, content, prefix=None, suffix=None):
        if prefix is None:
            prefix = self.get_form_property(field, "form_input_prefix")
        if suffix is None:
            suffix = self.get_form_property(field, "form_input_suffix")
        if prefix or suffix:
            content = html.div(class_="input-group")[
                html.div(class_="input-group-prepend", if_=prefix is not None)[
                    html.span(class_="input-group-text")[prefix]
                ],
                content,
                html.div(class_="input-group-append", if_=suffix is not None)[
                    html.span(class_="input-group-text")[suffix]
                ]
            ]
        return content

    def render_widget(self, field, prefix=None, suffix=None, hidden=None, form=None):
        widget = field.field.widget
        is_hidden = widget.is_hidden if hidden is None else hidden
        if isinstance(widget, (forms.TextInput, forms.Textarea, forms.Select)):
            widget.attrs["class"] = html.add_css_class(widget.attrs.get("class"), self.input_class, self.widget_class)
            if form:
                widget.attrs['form'] = form
            content = str(field)
        else:
            css_class = ["form-widget", html.widget_css_class(field), self.widget_class]
            if is_hidden:
                css_class.append("form-widget-hidden")
            widget.attrs["class"] = html.add_css_class(widget.attrs.get("class"), self.input_class)
            if form:
                widget.attrs['form'] = form
            content = html.div(class_=css_class)[str(field)]
        content = self.wrap_widget(field, content, prefix, suffix)
        result = str(content)
        return Markup(result)

    def render_field(self, field, layout, **kwargs):
        widget = field.field.widget
        prefix = kwargs.pop("prefix", None)
        suffix = kwargs.pop("suffix", None)
        if widget.is_hidden:
            return self.render_widget(field, prefix=prefix, suffix=suffix, hidden=True)
        for k,v in kwargs.items():
            setattr(field, k, v)
        layout_class = "layout-%s" % layout
        css_classes = ["form-field", html.field_css_class(field), self.field_class, layout_class]
        if field.field.required:
            css_classes.append("required")
        if field.errors:
            css_classes.append(self.failure_class)
        help_text = title = None
        if layout == "inline" and field.help_text:
            title = field.help_text
        else:
            help_text = self.render_help(field)
        return html.div(class_=css_classes, title=title)[
                self.render_label(field),
                self.render_widget(field, prefix=prefix, suffix=suffix),
                help_text,
                self.render_errors(field)
            ]


class BooleanInputMixin:

    def render_subwidget_wrapper(self, content):
        el = super(BooleanInputMixin, self).render_subwidget_wrapper(content)
        el = el(class_="form-check")
        return el

    def render_subwidget(self, code, label, selected, attrs):
        attrs["class"] = html.add_css_class(attrs.get("class", ""), "form-check-input")
        return super(BooleanInputMixin, self).render_subwidget(code, label, selected, attrs)


class RadioSelect(BooleanInputMixin, BootstrapWidget, ChoiceWidgetMixin, forms.RadioSelect):
    input_type = "radio"
    input_class = ""


class CheckboxSelectMultiple(BooleanInputMixin, BootstrapWidget, ChoiceWidgetMixin, forms.CheckboxSelectMultiple):
    input_type = "checkbox"
    input_class = ""


class CheckboxInput(BootstrapWidget, forms.CheckboxInput):

    field_class = "form-group form-check"
    input_class = ""

    def render_widget(self, field, **kwargs):
        if getattr(field, 'layout', None) == 'inline':
            return "%s %s" % (self.render_label(field), str(field))
        else:
            field.field.widget.attrs['class'] = html.add_css_class(field.field.widget.attrs.get("class", ""), "form-check-input")
            return html.div(class_="checkbox")[
                str(field),
                html.label(class_=self.label_class, for_=field.id_for_label)[
                    field.label
                ]
            ]


# class Select(BootstrapWidget, forms.Select):
#     input_class = "form-control bootstrap-select-control"


class SelectMultiple(BootstrapWidget, forms.SelectMultiple):
    input_class = "form-control bootstrap-select-control"


# class ClearableFileInput(BootstrapWidget, forms.ClearableFileInput):
#     widget_class = "custom-file"
#     input_class = "custom-file-input"
#     label_class = "custom-file-label"
#
#     def render_widget(self, field):
#         widget = field.field.widget
#         css_class = ["form-widget", html.widget_css_class(field), self.widget_class]
#         widget.attrs["class"] = html.add_css_class(widget.attrs.get("class"), self.input_class)
#         content = html.div(class_=css_class)[str(field), self.render_label(field)]
#         result = str(content)
#         return result


class SwitchInput(CheckboxInput):
    extra_attrs = {"class": "switch-input"}

    def render_widget(self, field):
        return "%s%s" % (self.render_label(field), str(field))


class DateRangeWidget(forms.MultiWidget):
    pass


class DateInput(BootstrapWidget, forms.DateInput):
    input_type = "date"


class TimeInput(BootstrapWidget, forms.TimeInput):
    input_type = "time"


class DateTimeInput(BootstrapWidget, forms.DateTimeInput):
    input_type = "datetime-local"
    extra_attrs = {"class": "datetimepicker"}


class NumberRangeWidget(forms.NumberInput):
    pass





from djingles.bootstrap4 import forms, models
from django.views import generic


class VerticalFormView(generic.FormView):
    template_name = "bootstrap4/form.html"
    form_class = forms.VerticalForm


class InlineFormView(generic.FormView):
    template_name = "bootstrap4/inline_form.html"
    form_class = forms.InlineForm


class ObjectTable(generic.ListView):
    pass


class ObjectDetailView(generic.DetailView):
    pass


class ObjectEditView(generic.UpdateView):
    pass


class ObjectDeleteView(generic.DeleteView):
    pass


class ObjectCreateView(generic.CreateView):
    pass



from django import forms
from collections import deque
import itertools
from djingles import html
from django.utils.encoding import force_text


class AbstractThemedWidget(object):
    attr_names = ("attrs", "choices")
    extra_attrs = {}

    @classmethod
    def __cached_subclasses(cls):
        cache_name = "_%s__widget_classes" % cls.__name__
        result = getattr(cls, cache_name, None)
        if result is None:
            result = {}
            subs = deque([cls])
            while subs:
                child = subs.popleft()
                result[child.__name__] = child
                subs.extendleft(child.__subclasses__())
            setattr(cls, cache_name, result)
        return result

    @classmethod
    def __mutate(cls, field):
        subclasses = cls.__cached_subclasses()
        widget = field.field.widget
        widget_class = widget.__class__
        name = widget_class.__name__
        if hasattr(widget, 'render_field'):
            return widget
        if name in subclasses:
            child = subclasses[name]
            widget = child.__consume(field)
        return widget

    @classmethod
    def __consume(cls, field):
        widget = field.field.widget
        common = cls.attr_names
        init_args = {}
        for name in common:
            try:
                init_args[name] = getattr(widget, name)
            except AttributeError:
                pass
        widget = cls(**init_args)
        field.field.widget = widget
        return widget

    @classmethod
    def widget_to_html(cls, field, **kwargs):
        widget = cls.__mutate(field)
        if hasattr(widget, 'render_widget'):
            return widget.render_widget(field, **kwargs)
        else:
            instance = cls()
            return instance.render_widget(field, **kwargs)

    @classmethod
    def field_to_html(cls, field, layout, **kwargs):
        field.layout = layout
        original_widget = field.field.widget
        widget = cls.__mutate(field)
        if hasattr(widget, 'render_field'):
            return widget.render_field(field, layout, **kwargs)
        else:
            instance = cls()
            return instance.render_field(field, layout, **kwargs)

    @classmethod
    def form_attrs(cls, form, **kwargs):
        return kwargs

    def render_field(self, field, layout, **kwargs):
        for k in kwargs:
            setattr(field.field, k, kwargs[k])
        return str(field)

    def render_widget(self, field):
        return str(field)

    def build_attrs(self, *args, **kwargs):
        attrs = super(AbstractThemedWidget, self).build_attrs(*args, **kwargs)
        extra = getattr(self, 'extra_attrs', {}).copy()
        css = html.CssClassList()
        css.append(extra.pop("class", ""))
        css.append(attrs.pop("class", ""))
        attrs.update(extra)
        if css:
            attrs['class'] = str(css)
        return attrs

    def add_css_class(self, classes):
        css = html.CssClassList()
        css.append(self.attrs.pop("class", ""))
        css.append(classes)
        self.attrs['class'] = str(css)


class ChoiceWidgetMixin(object):

    input_type = "radio"

    def render_from_field(self, field):
        return field.as_widget(widget=self)

    def render_subwidget_wrapper(self, content):
        classes = []
        return html.li(class_=classes)[content]

    def render_subwidget(self, code, label, selected, attrs):
        return [
            html.input(type=self.input_type, checked=selected, value=code, **attrs),
            html.label(for_=attrs.get("id"))[label]
        ]

    def render_wrapper(self, content, attrs):
        return html.ul(**attrs)[content].render()

    def render(self, name, value, attrs=None, choices=(), **kwargs):
        if value is None:
            value = ''
        attrs = self.build_attrs(attrs, self.attrs)
        children = []
        id_ = attrs.get("id")
        for i, (code, label) in enumerate(itertools.chain(self.choices, choices)):
            selected = html.is_selected_choice(value, code)
            child_attrs = attrs.copy()
            child_attrs.pop("class", None)
            if id_:
                child_attrs['id'] = "%s-%s" % (id_, i)
            child = self.render_subwidget(code, label, selected, child_attrs)
            children.append(self.render_subwidget_wrapper(child))
        return self.render_wrapper("".join(force_text(child) for child in children), attrs)






import re
import warnings
import mimetypes
# import urllib2
from collections import OrderedDict

from django.db.models.query_utils import Q
from django.utils.encoding import force_text
import os
from urllib.request import build_opener
from urllib.parse import urlparse
from django import forms
from django.core.validators import URLValidator
from django.core.files.uploadedfile import SimpleUploadedFile

from djingles import utils, html
from djingles.formatters.base import Formatter
from djingles.utils import feet_inches_to_cm, cm_to_feet_inches


__all__ = ["FileOrUrlInput", "HeightField", "HeightWidget", "SortField", "TableSortField"]


class FileOrUrlInput(forms.ClearableFileInput):
    
    def download_url(self, name, url):
        validate = URLValidator()
        try:
            validate(url)
        except forms.ValidationError as _:
            raise
            return None
        
        parsed_url = urlparse(url)
        path = parsed_url[2].strip("/")
        name = os.path.basename(path)
        opener = build_opener()
        opener.addheaders = [('User-agent', 'Mozilla/5.0')]
        ze_file = opener.open(url).read()
        file_obj = SimpleUploadedFile(name=name, content=ze_file, content_type=mimetypes.guess_type(name))
        file_obj.url = url
        return file_obj

    def value_from_datadict(self, data, files, name):
        if name in data and name not in files:
            url = forms.HiddenInput().value_from_datadict(data, files, name)
            result = self.download_url(name, url) if url and isinstance(url, str) else None
            files = files.copy() if files else {}
            files[name] = result
        return super(FileOrUrlInput, self).value_from_datadict(data, files, name)


class HeightWidget(forms.MultiWidget):

    def __init__(self, *args, **kwargs):
        widgets = [forms.TextInput(attrs={'placeholder': '5', 'size': '3'}), forms.TextInput(attrs={'placeholder': '6',
                                                                                                    'size': '3'})]
        super(HeightWidget,self).__init__(widgets, *args, **kwargs)

    def decompress(self, value):
        if value:
            result = cm_to_feet_inches(value)
            return result
        else:
            return [None,None]

    def format_output(self, rendered_widgets):
        return "%s ft   %s inches" % tuple(rendered_widgets)


class HeightField(forms.MultiValueField):
    widget = HeightWidget

    def __init__(self, *args, **kwargs):
        kwargs.pop('min_value',None)
        errors = self.default_error_messages.copy()
        if 'error_messages' in kwargs:
            errors.update(kwargs['error_messages'])
        reqd = kwargs.setdefault('required', False)
        fields = (
            forms.IntegerField(min_value=0,required=reqd),
            forms.IntegerField(min_value=0,required=reqd),
        )
        super(HeightField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        if data_list and all(d is not None for d in data_list):
            feet, inches = data_list
            return feet_inches_to_cm(feet, inches)
        return None


class SortField(forms.ChoiceField):

    def __init__(self, choices=(), toggle=True, modifier=None, **kwargs):
        kwargs.setdefault("required", False)
        kwargs.setdefault("widget", forms.HiddenInput)
        super(SortField, self).__init__(choices=choices, **kwargs)
        self.modifier = modifier
        self.toggle = toggle
        self.set_choices(choices)

    def __deepcopy__(self, memo):
        result = super().__deepcopy__(memo)
        result.toggle = self.toggle
        result.modifier = self.modifier
        return result

    def set_choices(self, choices):
        field_map = {}
        new_choices = []
        for i, (value, label) in enumerate(choices):
            position = str(i)
            new_choices.append((position, label))
            field_map[position] = re.sub(r'\s+', ' ', value.strip())
        self.choices = tuple(new_choices)
        self.field_map = field_map

    def valid_value(self, value):
        "Check to see if the provided value is a valid choice"
        text_value = force_text(value)
        if text_value.startswith("-"):
            text_value = text_value[1:]
        return text_value in self.field_map or super(SortField, self).valid_value(text_value)

    def build_links(self, request, bound_field):
        value = bound_field.value()
        field_name = bound_field.name
        text_value = force_text(value) if value is not None else None
        for k, v in self.choices:
            content = force_text(v)
            key = force_text(k)
            is_active = text_value and text_value == key
            if is_active and self.toggle:
                next_value = key if text_value.startswith("-") else "-%s" % key
            else:
                next_value = key
            url = utils.url_query_update(request.build_absolute_uri(), {field_name: next_value})
            yield html.Link(url=url, content=content, is_active=is_active)

    def invert_sign(self, name, neg):
        if name.startswith("-"):
            neg = not neg
        return "%s%s" % ("-" if neg else "", name.lstrip("-"))

    def filter_queryset(self, queryset, key, bound_field):
        neg = key.startswith("-")
        value = self.field_map[key.lstrip("-")]
        invert = lambda a: self.invert_sign(a, neg)
        values = map(invert, value.split())
        return queryset.order_by(*values)

    def get_value_for_name(self, name):
        for value, key in self.field_map.items():
            if name == key:
                return value


class TableSortField(SortField):

    def bind_form(self, form):
        formatter_class = form.context.get('formatter_class')
        if formatter_class is None:
            return
        column_dict = OrderedDict(formatter_class.base_formatters)
        choices = [(name, col.label or name.title()) for name, col in column_dict.items()
                        if col.sortable]
        self.column_dict = column_dict
        self.set_choices(choices)

    def filter_queryset(self, queryset, value, bound_field):
        print(">>>>>>>>>>>>>", value)
        text_value = force_text(value) if value is not None else None
        if not text_value:
            return queryset
        reverse = text_value.startswith("-")
        column_dict = self.column_dict
        name = text_value[1:] if reverse else text_value
        name = self.field_map[name]
        col = column_dict[name]
        if not col.sortable:
            return queryset
        attr = col.order_by or col.attr or name
        if col.reverse:
            reverse = not reverse
        if self.modifier:
            field = utils.get_related_field(queryset.model, attr)
            attr = self.modifier(attr, reverse, field)
        elif reverse:
            attr = "-%s" % attr
        result = queryset.order_by(attr)
        return result


class SearchField(forms.CharField):

    def __init__(self, field_names=None, *args, **kwargs):
        self.search_fields = field_names
        kwargs.setdefault("required", False)
        super(SearchField, self).__init__(*args, **kwargs)

    def filter_queryset(self, queryset, value, bound_field):
        q = Q()
        search_fields = self.search_fields
        if search_fields is None:
            meta = queryset.model._meta
            from django.db.models import TextField, CharField
            search_fields = [f.name for f in meta.fields if isinstance(f, (TextField, CharField))]
        for f in search_fields:
            op = "icontains"
            if f.startswith("="):
                f = f[1:]
                op = "iexact"
            key = "%s__%s" % (f, op)
            q |= Q(**{key: value})
        return queryset.filter(q).all()





from importlib import import_module
import pkgutil
from django.utils.module_loading import module_has_submodule
import os
import threading
import re
import hashlib
from collections import deque
import inspect
import base64
import pickle
from django.utils.encoding import force_bytes
from django.utils import timezone
from django.db import models
from django.apps import apps
from urllib import parse


first_cap_re = re.compile('(.)([A-Z][a-z]+)')

all_cap_re = re.compile('([a-z0-9])([A-Z])')



def camel_to_underscore(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()


def camel_to_hyphen(name):
    s1 = first_cap_re.sub(r'\1-\2', name)
    return all_cap_re.sub(r'\1-\2', s1).lower()


def join_with_underscore(*args):
    return '_'.join( str(a) for a in args if a )

def hyphen_to_camel(value):
    return "".join([x.title() for x in value.split('-')])

def underscore_to_camel(value):
    return "".join([x.title() for x in value.split('_')])


def page_number_range(index, limit, total):
    """
    :param index: Current page number
    :param limit: Maximum number of pages
    :param total: Total number of pages
    :return: iterator of int
    """
    index = max(min(total, index), 1)
    head = int((limit-1)/2)
    tail = limit - head
    first = max(index-head, 1)
    last = min(first+limit-1, total)
    if index+tail > last:
        first -= index+tail-last-1
    return range(max(first, 1), last+1)


def url_query_update(url, values, append=False):
    parts = parse.urlsplit(url)
    params = parse.parse_qs(parts.query)
    for key, value in values.items():
        if key not in params or not append:
            params[key] = value if isinstance(value, (list, tuple)) else [value]
        else:
            if isinstance(value, (list, tuple)):
                params[key].extend(value)
            else:
                params[key].append(value)
    return parts._replace(query=parse.urlencode(params, True)).geturl()


def qualified_name(cls):
    """
    Return fully-qualified name for classes and functions. Doesn't differentiate between
    bound and unbound methods
    :param cls:
    :return: str fully qualified name
    """
    # if hasattr(cls, "__qualname__"):
    #     return getattr(cls,"__qualname__")
    if inspect.ismodule(cls):
        return cls.__name__
    parts = [cls.__name__]
    if inspect.ismethod(cls):
        parts.append(cls.im_class.__name__)
    parts.append(cls.__module__)
    return ".".join(reversed(parts))


def create_hash(*args):
    """
    Returns a string representation of the provided arguments.
    Intended for use in other functions
    :param args: list of python objects
    :return: str
    """
    md = hashlib.md5()
    for a in args:
        md.update(repr(a).encode('utf-8'))
    return md.hexdigest()


def get_form_name(form_class_or_obj):
    form_class = form_class_or_obj if inspect.isclass(form_class_or_obj) else form_class_or_obj.__class__
    name = form_class.__qual_name__
    return create_hash(name)


def auth_login(request,user):
    from django.contrib.auth import login
    user.backend = 'django.contrib.auth.backends.ModelBackend'
    login(request,user)
    return user


def get_age(born):
    today = timezone.now().date()
    if not born: return 0
    try: # raised when birth date is February 29 and the current year is not a leap year
        birthday = born.replace(year=today.year)
    except ValueError:
        birthday = born.replace(year=today.year, day=born.day-1)
    if birthday > today:
        return today.year - born.year - 1
    else:
        return today.year - born.year


def get_request_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def get_request_latlng(request):
    from django.contrib.gis.geoip import GeoIP
    g = GeoIP()
    ip = get_request_ip(request)
    return g.lat_lon(ip)


def model_from_dict(model, kwargs, many_to_many=False):
    instance = model()
    model_update_from_dict(instance, kwargs, many_to_many=many_to_many)
    return instance


def model_to_dict(instance, fields=None, exclude=None):
    from django.db.models.fields.related import ManyToManyField, ForeignKey
    opts = instance._meta
    data = {}
    for f in opts.concrete_fields + opts.virtual_fields + opts.many_to_many:
        if fields and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue
        if isinstance(f, ForeignKey):
            data[f.get_attname()] = f.value_from_object(instance)
        elif isinstance(f, ManyToManyField):
            # If the object doesn't have a primary key yet, just use an empty
            # list for its m2m fields. Calling f.value_from_object will raise
            # an exception.
            if instance.pk is None:
                data[f.name] = []
            else:
                # MultipleChoiceWidget needs a list of pks, not object instances.
                qs = f.value_from_object(instance)
                if qs._result_cache is not None:
                    data[f.name] = [item.pk for item in qs]
                else:
                    data[f.name] = list(qs.values_list('pk', flat=True))
        else:
            data[f.name] = f.value_from_object(instance)
    return data


def model_update_from_dict(instance, kwargs, many_to_many=False):
    meta = instance.__class__._meta
    names = set(f.name for f in meta.get_fields())
    if not many_to_many:
        names.difference_update({a.name for a in meta.many_to_many})
    for k in names:
        if k in kwargs:
            setattr(instance, k, kwargs[k])


def update_object(instance, data):
    opts = instance._meta
    file_field_list = []
    # Wrap up the saving of m2m data as a function.
    def save_m2m():
        for f in opts.many_to_many:
            if f.name in data:
                f.save_form_data(instance, data[f.name])

    for f in opts.fields:
        if f.name not in data or isinstance(f, models.AutoField):
            continue
        if isinstance(f, models.FileField):
            file_field_list.append(f)
        else:
            f.save_form_data(instance, data[f.name])

    for f in file_field_list:
        f.save_form_data(instance, data[f.name])

    return save_m2m


def base64pickle_loads(data):
    encoded_data = base64.b64decode(force_bytes(data))
    return pickle.loads(encoded_data)


def base64pickle_dumps(data):
    serialized = pickle.dumps(data)
    return base64.b64encode(serialized).decode("ascii")


def iter_app_modules(module_name, deep=False):
    for config in apps.get_app_configs():
        package = config.module
        if module_has_submodule(package, module_name):
            full_name = "%s.%s" %(package.__name__, module_name)
            module = import_module(full_name)
            if not deep:
                yield module
            else:
                try:
                    pkg_path = module.__path__
                except AttributeError:
                    yield module
                else:
                    for importer, modname, ispkg in pkgutil.iter_modules(pkg_path):
                        if not ispkg:
                            yield importer.find_module(modname).load_module(modname)


def cm_to_feet_inches(value):
    v = float(value)
    feet = int(round(v / 30.48, 1))
    inches = int(round((v - feet * 30.48) / 2.54, 0))
    return feet, inches


def feet_inches_to_cm(feet, inches):
    result = int(round((feet * 12 + inches) * 2.54, 0))
    return result


def flatten(values):
    stack = deque(values)
    while stack:
        item = stack.popleft()
        if isinstance(item, (list, tuple)):
            stack.extendleft(reversed(item))
        else:
            yield item


def get_related_field(model, field_name):
    parts = field_name.split("__")
    field_name = parts.pop(0)
    field = model._meta.get_field(field_name)
    while parts and field.model:
        model = field.related_model
        item = parts.pop(0)
        field = model._meta.get_field(item)
    return field





import io
import ast
import re
from os import path
from setuptools import setup, find_packages

base_dir = path.abspath(path.dirname(__file__))

with io.open(path.join(base_dir, "README.md")) as f:
    long_description = f.read()


with io.open(path.join(base_dir, "djingles/_version.py")) as f:
    ctx = {}
    exec(f.read(), None, ctx)
    version = ctx['__version__']


setup(
    name="djingles",
    description="Set of django utilities",
    long_description=long_description,
    version=version,
    url="https://github.com/vivsh/djingles",
    license="BSD",
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: HTML'
    ],
    keywords="django utilities",
    install_requires=[
        "python-dateutil",
        "jinja2>=2.8",
        "django>=2",
        "click",
        "djangorestframework",
        "cookiecutter",
    ],
    python_requires='>=3',
    include_package_data=True,
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "djingles = djingles.commands.main:cli"
        ]
    }
)




from django.utils.encoding import force_text
import re
import json
from jinja2 import Markup


__all__ = ['html_json', 'html_attrs', "Element", "CssClassList", "CssStyle", 'add_css_class', 'empty']


def html_json(values):
    content = json.dumps(values)
    try:
        content = content.encode("unicode-escape")
    except LookupError:
        content = content.encode("string-escape")
    return Markup(content)


def html_attrs(*args, **kwargs):
    attr = HtmlAttr()
    attr.update(*args, **kwargs)
    return str(attr)


def add_css_class(original_class, *css_classes):
    css = CssClassList()
    css.append(original_class)
    css.append(css_classes)
    return str(css)


class CssClassList(object):

    def __init__(self):
        self.classes = []

    def __iter__(self):
        return iter(self.classes)

    def __len__(self):
        return len(self.classes)

    def copy(self):
        value = CssClassList()
        value.classes.extend(self.classes)
        return value

    def append(self, value):
        if isinstance(value, str):
            value = re.sub(r'\s+', ' ', value.strip())
            if len(value) == 1:
                value = value[0]
        if isinstance(value, (tuple, list)):
            for val in value:
                self.append(val)
        else:
            if value not in self.classes:
                self.classes.append(value)

    def __contains__(self, item):
        return item in self.classes

    def __str__(self):
        return " ".join(str(c) for c in self.classes if c)


class CssStyle(dict):

    def render(self):
        return ";".join("%s:%s" % (key.replace("_", "-"), value) for (key, value) in self.items())

    def __str__(self):
        return self.render()

    def copy(self):
        return CssStyle(super(CssStyle, self).copy())


def _normalize(key):
    if key.endswith("_"):
        key = key[:-1]
    key = key.replace("__", ":").replace("_", "-")
    return key


class HtmlAttr(object):

    def __init__(self):
        self.attrs = {}
        self.styles = CssStyle()
        self.classes = CssClassList()

    def copy(self):
        attr = HtmlAttr()
        attr.attrs = self.attrs.copy()
        attr.styles = self.styles.copy()
        attr.classes = self.classes.copy()
        return attr

    def dict(self):
        return dict(self)

    def __setitem__(self, key, value):
        self.set(key, value)

    def __getitem__(self, item):
        return dict(self)[item]

    def __len__(self):
        return len(dict(self))

    def get(self, key):
        return dict(self).get(key)

    def set(self, key, value):
        key = _normalize(key)
        if key in {"class"}:
            self.classes.append(value)
        elif key == "style":
            if isinstance(value, str):
                result = {}
                pairs = value.split(";")
                for p in pairs:
                    k, v = p.split(":", 1)
                    result[k] = v
                value = result
            self.styles.update(value)
        else:
            self.attrs[key] = value

    def update(self, *args, **attrs):
        values = {}
        values.update(*args, **attrs)
        for k, v in values.items():
            self.set(k, v)

    def __iter__(self):
        for k, v in self.attrs.items():
            yield k, v
        if self.classes:
            yield "class", str(self.classes)
        if self.styles:
            yield "style", self.styles.render()

    def render(self):
        pairs = []
        for key, value in self:
            if value is None or value is False:
                continue
            if value is True:
                pairs.append(key)
            else:
                if not isinstance(value, (str, bytes)):
                    value = html_json(value)
                pairs.append("%s='%s'" % (key, str(value)))
        return " ".join(pairs)

    def __str__(self):
        return self.render()


class Element(object):

    def __init__(self, tag):
        self.tag = tag
        self.attrib = HtmlAttr()
        self.children = []

    def __call__(self, **kwargs):
        el = self.copy()
        el.attrib.update(kwargs)
        return el

    def __getitem__(self, item):
        el = self.copy()
        if not isinstance(item, (list, tuple)):
            item = [item]
        for c in item:
            el.append(c)
        return el

    def copy(self):
        el = self.__class__(self.tag)
        el.attrib = self.attrib.copy()
        el.children = self.children[:]
        return el

    def mutate(self, tag):
        el = tag.copy()
        el.attrib.update(self.attrib.copy())
        el.children = self.children[:]
        return el

    def append(self, child):
        if child is None:
            return
        if isinstance(child, (list, tuple)):
            for c in child:
                self.append(c)
        else:
            self.children.append(child)

    def convert_to_text(self, el, *args, **kwargs):
        return el.render(*args, **kwargs) if hasattr(el, 'render') else force_text(el)

    def render_children(self, *args, **kwargs):
        return "".join(filter(None, (self.convert_to_text(c, *args, **kwargs)for c in self.children)))

    def render(self, ctx=None):
        if self.attrib.get('if') is False:
            return None
        attrs = self.attrib
        content = self.render_children(ctx)
        tag = _normalize(self.tag)
        return u"<{tag} {attrs}>{content}</{tag}>".format(**locals())

    def __str__(self):
        return self.render()

    def __html__(self):
        return self.render()


class Empty(Element):
    def render(self, *args, **kwargs):
        return self.render_children(*args, **kwargs)

empty = Empty("none")

for name in "html body link meta div span form section article aside main ul li ol dl dd dt p a strong "\
            "i fieldset legend b em input select button label nav textarea " \
            "table tbody tfoot thead tr td th figure caption img".split(" "):
    __all__.append(name)
    globals()[name] = Element(name)


if __name__ == '__main__':
    print(input(type="radio", checked=False).render())